/*
 * Copyright 2014, ACKme Networks
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of ACKme Networks;
 * the contents of this file may not be disclosed to third parties, copied
 * or duplicated in any form, in whole or in part, without the prior
 * written permission of ACKme Networks.
 */

#include "Wiconnect.h"
#include "internal/common.h"

/*************************************************************************************************/
WiconnectResult NetworkInterface::join(const char* ssid, const char *password, const Callback &completeHandler_)
{
    WiconnectResult result = WICONNECT_ERROR;

    enum
    {
        FS_SET_SSID,
        FS_SET_PASSWORD,
        FS_NETWORK_UP
    };

    CHECK_CALLBACK_AVAILABLE(completeHandler_);
    CHECK_OTHER_COMMAND_EXECUTING();

    if(wiconnect->internalProcessingState == FS_SET_SSID)
    {
        if(ssid == NULL ||
           WICONNECT_SUCCEEDED(result, wiconnect->sendCommand("set wlan.ssid %s", ssid)))
        {
            wiconnect->internalProcessingState = FS_SET_PASSWORD;
        }
    }

    if(wiconnect->internalProcessingState == FS_SET_PASSWORD)
    {
        if(password == NULL ||
           WICONNECT_SUCCEEDED(result, wiconnect->sendCommand("set wlan.passkey %s", password)))
        {
            wiconnect->internalProcessingState = FS_NETWORK_UP;
        }
    }

    if(wiconnect->internalProcessingState == FS_NETWORK_UP)
    {
        if(WICONNECT_SUCCEEDED(result, wiconnect->sendCommand("network_up")))
        {
#ifdef WICONNECT_ASYNC_TIMER_ENABLED
            if(completeHandler_.isValid())
            {
                if(monitorTimer.isRunning())
                {
                    result = WICONNECT_MONITOR_NOT_AVAILABLE;
                }
                else
                {
                    completeHandler = completeHandler_;
                    monitorTimer.start(this, &NetworkInterface::joinStatusMonitor, 1000);
                }
            }
#endif
        }
    }

    CHECK_CLEANUP_COMMAND();

    return result;
}



/*************************************************************************************************/
WiconnectResult NetworkInterface::leave()
{
    WiconnectResult result;

    CHECK_OTHER_COMMAND_EXECUTING();

#ifdef WICONNECT_ASYNC_TIMER_ENABLED
    monitorTimer.stop();
#endif
    result = wiconnect->sendCommand("network_down");

    CHECK_CLEANUP_COMMAND();

    return result;
}


/*************************************************************************************************/
WiconnectResult NetworkInterface::getNetworkStatus(NetworkStatus *statusPtr)
{
    WiconnectResult result;

    CHECK_OTHER_COMMAND_EXECUTING();

    if(WICONNECT_SUCCEEDED(result, wiconnect->sendCommand("get network.status")))
    {
        int32_t status;
        if(!WICONNECT_FAILED(result, wiconnect->responseToInt32(&status)))
        {
            *statusPtr = (NetworkStatus)status;
        }
    }

    CHECK_CLEANUP_COMMAND();

    return result;
}


#ifdef WICONNECT_ASYNC_TIMER_ENABLED

/*************************************************************************************************/
// this is called every 1s by the monitorTimer
void NetworkInterface::joinStatusMonitor()
{
    static char responseBuffer[4];
    static uint8_t cmdBuffer[sizeof(QueuedCommand)];
    QueuedCommand *cmd = (QueuedCommand*)cmdBuffer;

    monitorTimer.stop();

   *cmd = QueuedCommand(sizeof(responseBuffer), responseBuffer, "get network.status");

    wiconnect->enqueueCommand(cmd, Callback(this, &NetworkInterface::joinStatusCheckCallback));
}

/*************************************************************************************************/
// this is called on the completion of the 'get'network.status' command above
void NetworkInterface::joinStatusCheckCallback(WiconnectResult result, void *arg1, void *arg2)
{
    bool isComplete = true;

    QueuedCommand *cmd = (QueuedCommand*)arg1;

    if(result == WICONNECT_SUCCESS)
    {
        int32_t status;
        if(!StringUtil::strToInt32(cmd->responseBuffer, &status))
        {
            result = WICONNECT_RESPONSE_PARSE_ERROR;
        }
        else if(status == 0)
        {
            isComplete = false;
        }
    }

    if(isComplete)
    {
        completeHandler.call(result, NULL, NULL);
    }
    else
    {
        monitorTimer.start(this, &NetworkInterface::joinStatusMonitor, 1000);
    }
}

#endif
